/*
    VTun - Virtual Tunnel over TCP/IP network.

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * sniff_dev.c,v 1.0 2003/04/06 22:10:00 silvio & Vadim Dvorkin
 *             v 1.1 2003/14/06 15:32:00 silvio & Vadim Dvorkin
 *             v 1.2 2003/18/06 12:20:09 silvio & vadim 
 *                            (return in write and memset in filter fixed)
 */

#include "config.h"

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>

#include "vtun.h"
#include "lib.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "sniff_dev.h"
#include "linkfd.h"

extern struct vtun_host *lfd_host;

struct ifreq    ifr;
struct pkt_head filter_mask[6]; /* two masks for each case + 0 + result */
int             masks_created = 0;

void            create_masks(void);
void           *struct_xor_and(int *, int *, int *, int, int *);

int sniff_open(char *dev)
{
    int             fd, rc;

    if (*dev)
        if ((fd = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
            perror("Can't open socket: ");
            return -1;
        }

    /* set promisc mode on device */
    strcpy(ifr.ifr_name, dev);
    if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
        perror("Can't get config flags: ");
        return -1;
    }
    ifr.ifr_flags |= IFF_PROMISC;

    if ((rc = ioctl(fd, SIOCSIFFLAGS, &ifr)) < 0
        || ioctl(fd, SIOCGIFHWADDR, &ifr) < 0) {
        perror("Can't set promisc / get hardware address: ");
        return -1;
    }

    vtun_syslog(LOG_INFO,
                "Hardware address found for %s (socket %d): %s\n",
                dev, fd, ether_ntoa((void *) ifr.ifr_hwaddr.sa_data));

    return fd;
}

int sniff_close(int fd, char *dev)
{
    /* get out of promisc mode */
    ifr.ifr_flags &= ~IFF_PROMISC;
    ioctl(fd, SIOCSIFFLAGS, &ifr);
    return close(fd);
}

int empty_write(int fd, char *buf, int len)
{
    return len;
}

int empty_read(int fd, char *buf, int len)
{
    read(fd, buf, len); /* no checks since we don't care */
    errno = EAGAIN;
    return -1;
}

/*
 * filter the frame by hwaddr, ip and port,
 * if the packet matches, then its an internal
 * vtun packet and should not be shown (return 0).
 */
int sniff_filter(pkt_head * buf)
{

    if (masks_created) {

        struct_xor_and((int *) &filter_mask[0], (int *) buf,
                       (int *) &filter_mask[1], sizeof(pkt_head),
                       (int *) &filter_mask[5]);

        if (!memcmp(&filter_mask[5], &filter_mask[4], sizeof(pkt_head)))
            return 0;

        struct_xor_and((int *) &filter_mask[2], (int *) buf,
                       (int *) &filter_mask[2], sizeof(pkt_head),
                       (int *) &filter_mask[5]);
        if (!memcmp(&filter_mask[5], &filter_mask[4], sizeof(pkt_head)))
            return 0;           /* no permit */
    } else
        create_masks();

    return 1;
}

int sniff_read(int fd, char *buf, int len)
{
    int             rc;

    if ((rc = read(fd, buf, len)) > 0) {
        if (sniff_filter((pkt_head *) buf)) {
            return rc;
        } else {
            errno = EAGAIN;
            return -1;
        }
    }

    return rc;
}

void create_masks(void)
{
    /* static mask creation */

    struct sockaddr_in local, remote;
    int             namelen = sizeof(struct sockaddr);

    memset(&(filter_mask[0]), 0, sizeof(pkt_head) * 6);

    getpeername(lfd_host->rmt_fd, (struct sockaddr *) &remote, &namelen);
    getsockname(lfd_host->rmt_fd, (struct sockaddr *) &local, &namelen);

    /* first mask = packets we send */

    memcpy(filter_mask[0].eth_head.h_source, ifr.ifr_hwaddr.sa_data,
           ETH_ALEN);
    memset(filter_mask[1].eth_head.h_source, 0xFF, ETH_ALEN);
    filter_mask[0].eth_head.h_proto = htons(ETH_P_IP);
    filter_mask[1].eth_head.h_proto = (unsigned short) -1;
    filter_mask[0].ip_head.saddr = local.sin_addr.s_addr;
    filter_mask[1].ip_head.saddr = (u_int32_t) - 1;
    filter_mask[0].ip_head.daddr = remote.sin_addr.s_addr;
    filter_mask[1].ip_head.daddr = (u_int32_t) - 1;
#ifdef __FAVOR_BSD
    filter_mask[0].tcp_head.th_sport = local.sin_port;
    filter_mask[1].tcp_head.th_sport = (u_int16_t) - 1;
    filter_mask[0].tcp_head.th_dport = remote.sin_port;
    filter_mask[1].tcp_head.th_dport = (u_int16_t) - 1;
#else
    filter_mask[0].tcp_head.source = local.sin_port;
    filter_mask[1].tcp_head.source = (u_int16_t) - 1;
    filter_mask[0].tcp_head.dest = remote.sin_port;
    filter_mask[1].tcp_head.dest = (u_int16_t) - 1;
#endif

    /* second mask = acks we receive */

    memcpy(filter_mask[2].eth_head.h_dest, ifr.ifr_hwaddr.sa_data,
           ETH_ALEN);
    memset(filter_mask[3].eth_head.h_dest, 0xFF, ETH_ALEN);
    filter_mask[2].eth_head.h_proto = htons(ETH_P_IP);
    filter_mask[3].eth_head.h_proto = (unsigned short) -1;
    filter_mask[2].ip_head.daddr = local.sin_addr.s_addr;
    filter_mask[3].ip_head.daddr = (u_int32_t) - 1;
    filter_mask[2].ip_head.saddr = remote.sin_addr.s_addr;
    filter_mask[3].ip_head.saddr = (u_int32_t) - 1;
#ifdef __FAVOR_BSD
    filter_mask[2].tcp_head.th_dport = local.sin_port;
    filter_mask[3].tcp_head.th_dport = (u_int16_t) - 1;
    filter_mask[2].tcp_head.th_sport = remote.sin_port;
    filter_mask[3].tcp_head.th_sport = (u_int16_t) - 1;
#else
    filter_mask[2].tcp_head.dest = local.sin_port;
    filter_mask[3].tcp_head.dest = (u_int16_t) - 1;
    filter_mask[2].tcp_head.source = remote.sin_port;
    filter_mask[3].tcp_head.source = (u_int16_t) - 1;
#endif

    masks_created = 1;
}

void           *struct_xor_and(int *buf, int *xor_mask, int *and_mask,
                               int len, int *result)
{
    /* for now handles only lens divisible by 4 */
    int             i;

    for (i = 0; i < len / 4; i++) {
        *result = (*buf ^ *xor_mask) & *and_mask;
        result++;
        buf++;
        xor_mask++;
        and_mask++;
    }

    return result;
}
